@@ -48,16 +48,22 @@
 #define UNIX_PATH_MAX    108
 #endif /* UNIX_PATH_MAX */
 
-#if defined(QB_LINUX) || defined(QB_SOLARIS)
-#define QB_SUN_LEN(a) sizeof(*(a))
+/*
+ * SUN_LEN() does a strlen() on sun_path, but if you are trying to use the
+ * "Linux abstract namespace" (you have to set sun_path[0] == '\0') then
+ * the strlen() doesn't work.
+ */
+#if defined(SUN_LEN)
+#define QB_SUN_LEN(a) ((a)->sun_path[0] == '\0') ? sizeof(*(a)) : SUN_LEN(a)
 #else
-#define QB_SUN_LEN(a) SUN_LEN(a)
+#define QB_SUN_LEN(a) sizeof(*(a))
 #endif
 
 struct ipc_us_control {
 	int32_t sent;
 	int32_t flow_control;
 };
+#define SHM_CONTROL_SIZE (3 * sizeof(struct ipc_us_control))
 
 struct ipc_auth_ugp {
 	uid_t uid;
@@ -68,14 +74,36 @@ struct ipc_auth_ugp {
 static int32_t qb_ipcs_us_connection_acceptor(int fd, int revent, void *data);
 static int32_t qb_ipc_us_fc_get(struct qb_ipc_one_way *one_way);
 
-#ifdef SO_NOSIGPIPE
 static void
 socket_nosigpipe(int32_t s)
 {
+#ifdef SO_NOSIGPIPE
 	int32_t on = 1;
 	setsockopt(s, SOL_SOCKET, SO_NOSIGPIPE, (void *)&on, sizeof(on));
+#endif /* SO_NOSIGPIPE */
+}
+
+
+enum qb_sigpipe_ctl {
+       QB_SIGPIPE_IGNORE,
+       QB_SIGPIPE_DEFAULT,
+};
+
+static void sigpipe_ctl(enum qb_sigpipe_ctl ctl)
+{
+#if !defined(MSG_NOSIGNAL) && !defined(SO_NOSIGPIPE)
+	struct sigaction act;
+	struct sigaction oact;
+
+	act.sa_handler = SIG_IGN;
+
+	if (ctl == QB_SIGPIPE_IGNORE) {
+		sigaction(SIGPIPE, &act, &oact);
+	} else {
+		sigaction(SIGPIPE, &oact, NULL);
+	}
+#endif  /* !MSG_NOSIGNAL && !defined(SO_NOSIGPIPE) */
 }
-#endif
 
 #ifndef MSG_NOSIGNAL
 #define MSG_NOSIGNAL 0
@@ -85,40 +113,35 @@ ssize_t
 qb_ipc_us_send(struct qb_ipc_one_way *one_way, const void *msg, size_t len)
 {
 	int32_t result;
-	struct msghdr msg_send;
-	struct iovec iov_send;
-	char *rbuf = (char *)msg;
 	int32_t processed = 0;
-	struct ipc_us_control *ctl = NULL;
+	char *rbuf = (char *)msg;
 
-	msg_send.msg_iov = &iov_send;
-	msg_send.msg_iovlen = 1;
-	msg_send.msg_name = 0;
-	msg_send.msg_namelen = 0;
-
-#if !defined(QB_SOLARIS)
-	msg_send.msg_control = 0;
-	msg_send.msg_controllen = 0;
-	msg_send.msg_flags = 0;
-#else
-	msg_send.msg_accrights = NULL;
-	msg_send.msg_accrightslen = 0;
-#endif
+	sigpipe_ctl(QB_SIGPIPE_IGNORE);
 
 retry_send:
-	iov_send.iov_base = &rbuf[processed];
-	iov_send.iov_len = len - processed;
+	result = send(one_way->u.us.sock,
+		      &rbuf[processed],
+		      len - processed,
+		      MSG_NOSIGNAL);
 
-	result = sendmsg(one_way->u.us.sock, &msg_send, MSG_NOSIGNAL);
 	if (result == -1) {
-		return -errno;
+		if (errno == EAGAIN && processed > 0) {
+			goto retry_send;
+		} else {
+			sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+			return -errno;
+		}
 	}
 
 	processed += result;
 	if (processed != len) {
 		goto retry_send;
 	}
+
+	sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+
 	if (one_way->type == QB_IPC_SOCKET) {
+		struct ipc_us_control *ctl = NULL;
 		ctl = (struct ipc_us_control *)one_way->u.us.shared_data;
 		if (ctl) {
 			qb_atomic_int_inc(&ctl->sent);
@@ -132,46 +155,52 @@ qb_ipc_us_sendv(struct qb_ipc_one_way *one_way, const struct iovec *iov,
 		size_t iov_len)
 {
 	int32_t result;
-	struct msghdr msg_send;
 	int32_t processed = 0;
-	size_t len = 0;
-	int32_t i;
-	struct ipc_us_control *ctl = NULL;
+	int32_t total_processed = 0;
+	int32_t iov_p = 0;
+	char *rbuf = (char *)iov[iov_p].iov_base;
 
-	for (i = 0; i < iov_len; i++) {
-		len += iov[i].iov_len;
-	}
-	msg_send.msg_iov = (struct iovec *)iov;
-	msg_send.msg_iovlen = iov_len;
-	msg_send.msg_name = 0;
-	msg_send.msg_namelen = 0;
-
-#if !defined(QB_SOLARIS)
-	msg_send.msg_control = 0;
-	msg_send.msg_controllen = 0;
-	msg_send.msg_flags = 0;
-#else
-	msg_send.msg_accrights = NULL;
-	msg_send.msg_accrightslen = 0;
-#endif
+	sigpipe_ctl(QB_SIGPIPE_IGNORE);
 
 retry_send:
-	result = sendmsg(one_way->u.us.sock, &msg_send, MSG_NOSIGNAL);
+	result = send(one_way->u.us.sock,
+		      &rbuf[processed],
+		      iov[iov_p].iov_len - processed,
+		      MSG_NOSIGNAL);
+
 	if (result == -1) {
-		return -errno;
+		if (errno == EAGAIN &&
+		    (processed > 0 || iov_p > 0)) {
+			goto retry_send;
+		} else {
+			sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+			return -errno;
+		}
 	}
 
 	processed += result;
-	if (processed != len) {
+	if (processed == iov[iov_p].iov_len) {
+		iov_p++;
+		total_processed += processed;
+		if (iov_p < iov_len) {
+			processed = 0;
+			rbuf = (char *)iov[iov_p].iov_base;
+			goto retry_send;
+		}
+	} else {
 		goto retry_send;
 	}
+
+	sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+
 	if (one_way->type == QB_IPC_SOCKET) {
+		struct ipc_us_control *ctl;
 		ctl = (struct ipc_us_control *)one_way->u.us.shared_data;
 		if (ctl) {
 			qb_atomic_int_inc(&ctl->sent);
 		}
 	}
-	return processed;
+	return total_processed;
 }
 
 static ssize_t
@@ -180,6 +209,8 @@ qb_ipc_us_recv_msghdr(int32_t s, struct msghdr *hdr, char *msg, size_t len)
 	int32_t result;
 	int32_t processed = 0;
 
+	sigpipe_ctl(QB_SIGPIPE_IGNORE);
+
 retry_recv:
 	hdr->msg_iov->iov_base = &msg[processed];
 	hdr->msg_iov->iov_len = len - processed;
@@ -189,34 +220,50 @@ retry_recv:
 		goto retry_recv;
 	}
 	if (result == -1) {
+		sigpipe_ctl(QB_SIGPIPE_DEFAULT);
 		return -errno;
 	}
-#if defined(QB_SOLARIS) || defined(QB_BSD) || defined(QB_DARWIN)
-	/* On many OS poll never return POLLHUP or POLLERR.
-	 * EOF is detected when recvmsg return 0.
-	 */
 	if (result == 0) {
-		return -errno;	//ENOTCONN
+		sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+		qb_util_log(LOG_DEBUG,
+			    "recv(fd %d) got 0 bytes assuming ENOTCONN", s);
+		return -ENOTCONN;
 	}
-#endif
 
 	processed += result;
 	if (processed != len) {
 		goto retry_recv;
 	}
+	sigpipe_ctl(QB_SIGPIPE_DEFAULT);
 	assert(processed == len);
 
 	return processed;
 }
 
 int32_t
-qb_ipc_us_recv_ready(struct qb_ipc_one_way * one_way, int32_t ms_timeout)
+qb_ipc_us_sock_error_is_disconnected(int err)
+{
+	if (err == -EAGAIN ||
+	    err == -ETIMEDOUT ||
+	    err == -EINTR ||
+#ifdef EWOULDBLOCK
+	    err == -EWOULDBLOCK ||
+#endif
+	    err == -EINVAL) {
+		return QB_FALSE;
+	}
+	return QB_TRUE;
+}
+
+int32_t
+qb_ipc_us_ready(struct qb_ipc_one_way * one_way,
+		int32_t ms_timeout, int32_t events)
 {
 	struct pollfd ufds;
 	int32_t poll_events;
 
 	ufds.fd = one_way->u.us.sock;
-	ufds.events = POLLIN;
+	ufds.events = events;
 	ufds.revents = 0;
 
 	poll_events = poll(&ufds, 1, ms_timeout);
@@ -224,7 +271,14 @@ qb_ipc_us_recv_ready(struct qb_ipc_one_way * one_way, int32_t ms_timeout)
 		return -EAGAIN;
 	} else if (poll_events == -1) {
 		return -errno;
-	} else if (poll_events == 1 && (ufds.revents & (POLLERR | POLLHUP))) {
+	} else if (poll_events == 1 && (ufds.revents & POLLERR)) {
+		qb_util_log(LOG_DEBUG, "poll(fd %d) got POLLERR", one_way->u.us.sock);
+		return -ENOTCONN;
+	} else if (poll_events == 1 && (ufds.revents & POLLHUP)) {
+		qb_util_log(LOG_DEBUG, "poll(fd %d) got POLLHUP", one_way->u.us.sock);
+		return -ENOTCONN;
+	} else if (poll_events == 1 && (ufds.revents & POLLNVAL)) {
+		qb_util_log(LOG_DEBUG, "poll(fd %d) got POLLNVAL", one_way->u.us.sock);
 		return -ENOTCONN;
 	}
 	return 0;
@@ -241,27 +295,44 @@ qb_ipc_us_recv(struct qb_ipc_one_way * one_way,
 	int32_t processed = 0;
 	int32_t to_recv = len;
 	char *data = msg;
-	struct ipc_us_control *ctl = NULL;
+
+	sigpipe_ctl(QB_SIGPIPE_IGNORE);
 
 retry_recv:
-	result = recv(one_way->u.us.sock, &data[processed], to_recv, MSG_NOSIGNAL | MSG_WAITALL);
-	if (timeout == -1) {
-		if (result == -1 && errno == EAGAIN) {
+	result = recv(one_way->u.us.sock, &data[processed], to_recv,
+		      MSG_NOSIGNAL | MSG_WAITALL);
+
+	if (result == -1) {
+		if (errno == EAGAIN &&
+		    (processed > 0 || timeout == -1)) {
 			goto retry_recv;
+		} else if (errno == ECONNRESET || errno == EPIPE) {
+			sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+			qb_util_perror(LOG_DEBUG,
+				       "recv(fd %d) converting to ENOTCONN",
+				       one_way->u.us.sock);
+			return -ENOTCONN;
+		} else {
+			sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+			return -errno;
 		}
 	}
+
 	if (result == 0) {
+		sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+		qb_util_log(LOG_DEBUG,
+			    "recv(fd %d) got 0 bytes assuming ENOTCONN",
+			    one_way->u.us.sock);
 		return -ENOTCONN;
 	}
-	if (result == -1) {
-		return -errno;
-	}
 	processed += result;
 	to_recv -= result;
 	if (processed != len) {
 		goto retry_recv;
 	}
+	sigpipe_ctl(QB_SIGPIPE_DEFAULT);
 	if (one_way->type == QB_IPC_SOCKET) {
+		struct ipc_us_control *ctl = NULL;
 		ctl = (struct ipc_us_control *)one_way->u.us.shared_data;
 		if (ctl) {
 			(void)qb_atomic_int_dec_and_test(&ctl->sent);
@@ -271,34 +342,58 @@ retry_recv:
 }
 
 /*
- * recv a message of unknow size.
- * (could use MSG_PEEK here)
+ * recv a message of unknown size.
  */
 static ssize_t
 qb_ipc_us_recv_at_most(struct qb_ipc_one_way * one_way,
 	       void *msg, size_t len, int32_t timeout)
 {
 	int32_t result;
+	int32_t processed = 0;
+	int32_t to_recv = sizeof(struct qb_ipc_request_header);
+	char *data = msg;
 	struct ipc_us_control *ctl = NULL;
+	struct qb_ipc_request_header *hdr = NULL;
+
+	sigpipe_ctl(QB_SIGPIPE_IGNORE);
 
 retry_recv:
-	result = recv(one_way->u.us.sock, msg, len, MSG_NOSIGNAL | MSG_WAITALL);
-	if (timeout == -1) {
-		if (result == -1 && errno == EAGAIN) {
+	result = recv(one_way->u.us.sock, &data[processed], to_recv,
+		      MSG_NOSIGNAL | MSG_WAITALL);
+	if (result == -1) {
+		if (errno == EAGAIN &&
+		    (processed > 0 || timeout == -1)) {
 			goto retry_recv;
+		} else {
+			sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+			return -errno;
 		}
-	}
-	if (result == 0) {
+	} else if (result == 0) {
+		sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+		qb_util_log(LOG_DEBUG,
+			    "recv(fd %d) got 0 bytes assuming ENOTCONN",
+			    one_way->u.us.sock);
 		return -ENOTCONN;
 	}
-	if (result == -1) {
-		return -errno;
+
+	processed += result;
+	if (processed >= sizeof(struct qb_ipc_request_header) && hdr == NULL) {
+		hdr = (struct qb_ipc_request_header*)msg;
 	}
+	if (hdr) {
+		to_recv = hdr->size - processed;
+	} else {
+		to_recv = len - processed;
+	}
+	if (to_recv > 0) {
+		goto retry_recv;
+	}
+	sigpipe_ctl(QB_SIGPIPE_DEFAULT);
 	ctl = (struct ipc_us_control *)one_way->u.us.shared_data;
 	if (ctl) {
 		(void)qb_atomic_int_dec_and_test(&ctl->sent);
 	}
-	return result;
+	return processed;
 }
 
 
@@ -309,17 +404,13 @@ qb_ipcc_us_sock_connect(const char *socket_name, int32_t * sock_pt)
 	struct sockaddr_un address;
 	int32_t res = 0;
 
-#if defined(QB_SOLARIS)
 	request_fd = socket(PF_UNIX, SOCK_STREAM, 0);
-#else
-	request_fd = socket(PF_LOCAL, SOCK_STREAM, 0);
-#endif
 	if (request_fd == -1) {
 		return -errno;
 	}
-#ifdef SO_NOSIGPIPE
+
 	socket_nosigpipe(request_fd);
-#endif /* SO_NOSIGPIPE */
+
 	res = qb_sys_fd_nonblock_cloexec_set(request_fd);
 	if (res < 0) {
 		goto error_connect;
@@ -327,11 +418,11 @@ qb_ipcc_us_sock_connect(const char *socket_name, int32_t * sock_pt)
 
 	memset(&address, 0, sizeof(struct sockaddr_un));
 	address.sun_family = AF_UNIX;
-#if defined(QB_BSD) || defined(QB_DARWIN)
-	address.sun_len = SUN_LEN(&address);
+#ifdef HAVE_STRUCT_SOCKADDR_UN_SUN_LEN
+	address.sun_len = QB_SUN_LEN(&address);
 #endif
 
-#if defined(QB_LINUX)
+#if defined(QB_LINUX) || defined(QB_CYGWIN)
 	snprintf(address.sun_path + 1, UNIX_PATH_MAX - 1, "%s", socket_name);
 #else
 	snprintf(address.sun_path, UNIX_PATH_MAX, "%s/%s", SOCKETDIR,
@@ -366,12 +457,21 @@ qb_ipcc_us_setup_connect(struct qb_ipcc_connection *c,
 {
 	int32_t res;
 	struct qb_ipc_connection_request request;
+#ifdef QB_LINUX
+	int off = 0;
+	int on = 1;
+#endif
 
 	res = qb_ipcc_us_sock_connect(c->name, &c->setup.u.us.sock);
 	if (res != 0) {
 		return res;
 	}
 
+#ifdef QB_LINUX
+	setsockopt(c->setup.u.us.sock, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
+#endif
+
+	memset(&request, 0, sizeof(request));
 	request.hdr.id = QB_IPC_MSG_AUTHENTICATE;
 	request.hdr.size = sizeof(request);
 	request.max_msg_size = c->setup.max_msg_size;
@@ -381,6 +481,10 @@ qb_ipcc_us_setup_connect(struct qb_ipcc_connection *c,
 		return res;
 	}
 
+#ifdef QB_LINUX
+	setsockopt(c->setup.u.us.sock, SOL_SOCKET, SO_PASSCRED, &off, sizeof(off));
+#endif
+
 	res =
 	    qb_ipc_us_recv(&c->setup, r,
 			   sizeof(struct qb_ipc_connection_response), -1);
@@ -397,7 +501,7 @@ qb_ipcc_us_setup_connect(struct qb_ipcc_connection *c,
 static void
 qb_ipcc_us_disconnect(struct qb_ipcc_connection *c)
 {
-	munmap(c->request.u.us.shared_data, sizeof(struct ipc_us_control));
+	munmap(c->request.u.us.shared_data, SHM_CONTROL_SIZE);
 	unlink(c->request.u.us.shared_file_name);
 	close(c->request.u.us.sock);
 	close(c->event.u.us.sock);
@@ -411,7 +515,9 @@ qb_ipcc_us_connect(struct qb_ipcc_connection *c,
 	struct qb_ipc_event_connection_request request;
 	char path[PATH_MAX];
 	int32_t fd_hdr;
-	struct ipc_us_control *ctl;
+	char * shm_ptr;
+
+	qb_atomic_init();
 
 	c->needs_sock_for_poll = QB_FALSE;
 	c->funcs.send = qb_ipc_us_send;
@@ -425,27 +531,24 @@ qb_ipcc_us_connect(struct qb_ipcc_connection *c,
 	c->setup.u.us.sock = -1;
 
 	fd_hdr = qb_sys_mmap_file_open(path, r->request,
-				       sizeof(struct ipc_us_control), O_RDWR);
+				       SHM_CONTROL_SIZE, O_RDWR);
 	if (fd_hdr < 0) {
 		res = -errno;
 		qb_util_perror(LOG_ERR, "couldn't open file for mmap");
 		return res;
 	}
 	(void)strlcpy(c->request.u.us.shared_file_name, r->request, NAME_MAX);
-	c->request.u.us.shared_data = mmap(0,
-					   sizeof(struct ipc_us_control),
-					   PROT_READ | PROT_WRITE, MAP_SHARED,
-					   fd_hdr, 0);
+	shm_ptr = mmap(0, SHM_CONTROL_SIZE,
+		       PROT_READ | PROT_WRITE, MAP_SHARED, fd_hdr, 0);
 
-	if (c->request.u.us.shared_data == MAP_FAILED) {
+	if (shm_ptr == MAP_FAILED) {
 		res = -errno;
 		qb_util_perror(LOG_ERR, "couldn't create mmap for header");
 		goto cleanup_hdr;
 	}
-
-	ctl = (struct ipc_us_control *)c->request.u.us.shared_data;
-	ctl->sent = 0;
-	ctl->flow_control = 0;
+	c->request.u.us.shared_data = shm_ptr;
+	c->response.u.us.shared_data = shm_ptr + sizeof(struct ipc_us_control);
+	c->event.u.us.shared_data =  shm_ptr + (2 * sizeof(struct ipc_us_control));
 
 	close(fd_hdr);
 
@@ -454,13 +557,14 @@ qb_ipcc_us_connect(struct qb_ipcc_connection *c,
 		goto cleanup_hdr;
 	}
 
+	memset(&request, 0, sizeof(request));
 	request.hdr.id = QB_IPC_MSG_NEW_EVENT_SOCK;
 	request.hdr.size = sizeof(request);
 	request.connection = r->connection;
 	res = qb_ipc_us_send(&c->event, &request, request.hdr.size);
 	if (res < 0) {
 		qb_ipcc_us_sock_close(c->event.u.us.sock);
-		return res;
+		goto cleanup_hdr;
 	}
 
 	return 0;
@@ -468,7 +572,7 @@ qb_ipcc_us_connect(struct qb_ipcc_connection *c,
 cleanup_hdr:
 	close(fd_hdr);
 	unlink(r->request);
-	munmap(c->request.u.us.shared_data, sizeof(struct ipc_us_control));
+	munmap(c->request.u.us.shared_data, SHM_CONTROL_SIZE);
 	return res;
 }
 
@@ -486,11 +590,7 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 	/*
 	 * Create socket for IPC clients, name socket, listen for connections
 	 */
-#if defined(QB_SOLARIS)
 	s->server_sock = socket(PF_UNIX, SOCK_STREAM, 0);
-#else
-	s->server_sock = socket(PF_LOCAL, SOCK_STREAM, 0);
-#endif
 	if (s->server_sock == -1) {
 		res = -errno;
 		qb_util_perror(LOG_ERR, "Cannot create server socket");
@@ -509,7 +609,7 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 #endif
 
 	qb_util_log(LOG_INFO, "server name: %s", s->name);
-#if defined(QB_LINUX)
+#if defined(QB_LINUX) || defined(QB_CYGWIN)
 	snprintf(un_addr.sun_path + 1, UNIX_PATH_MAX - 1, "%s", s->name);
 #else
 	{
@@ -541,7 +641,7 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 	 * Allow everyone to write to the socket since the IPC layer handles
 	 * security automatically
 	 */
-#if !defined(QB_LINUX)
+#if !defined(QB_LINUX) && !defined(QB_CYGWIN)
 	res = chmod(un_addr.sun_path, S_IRWXU | S_IRWXG | S_IRWXO);
 #endif
 	if (listen(s->server_sock, SERVER_BACKLOG) == -1) {
@@ -585,7 +685,7 @@ handle_new_connection(struct qb_ipcs_service *s,
 		return -ENOMEM;
 	}
 
-	c->receive_buf = malloc(req->max_msg_size);
+	c->receive_buf = calloc(1, req->max_msg_size);
 	if (c->receive_buf == NULL) {
 		free(c);
 		qb_ipcc_us_sock_close(sock);
@@ -596,9 +696,13 @@ handle_new_connection(struct qb_ipcs_service *s,
 	c->response.max_msg_size = req->max_msg_size;
 	c->event.max_msg_size = req->max_msg_size;
 	c->pid = ugp->pid;
-	c->euid = ugp->uid;
-	c->egid = ugp->gid;
+	c->auth.uid = c->euid = ugp->uid;
+	c->auth.gid = c->egid = ugp->gid;
+	c->auth.mode = 0600;
 	c->stats.client_pid = ugp->pid;
+	snprintf(c->description, CONNECTION_DESCRIPTION,
+		 "%d-%d-%d", s->pid, ugp->pid,
+		 c->setup.u.us.sock);
 
 	if (auth_result == 0 && c->service->serv_fns.connection_accept) {
 		res = c->service->serv_fns.connection_accept(c,
@@ -608,7 +712,8 @@ handle_new_connection(struct qb_ipcs_service *s,
 		goto send_response;
 	}
 
-	qb_util_log(LOG_INFO, "IPC credentials authenticated");
+	qb_util_log(LOG_DEBUG, "IPC credentials authenticated (%s)",
+		    c->description);
 
 	memset(&response, 0, sizeof(response));
 	if (s->funcs.connect) {
@@ -630,19 +735,15 @@ handle_new_connection(struct qb_ipcs_service *s,
 					       POLLIN | POLLPRI | POLLNVAL,
 					       c,
 					       qb_ipcs_dispatch_connection_request);
+		if (res < 0) {
+			qb_util_log(LOG_ERR,
+				    "Error adding socket to mainloop (%s).",
+				    c->description);
+		}
 	}
 	if (s->type == QB_IPC_SOCKET) {
 		c->request.u.us.sock = c->setup.u.us.sock;
 		c->response.u.us.sock = c->setup.u.us.sock;
-		res = s->poll_fns.dispatch_add(s->poll_priority,
-					       c->request.u.us.sock,
-					       POLLIN | POLLPRI | POLLNVAL,
-					       c,
-					       qb_ipcs_dispatch_connection_request);
-		if (res < 0) {
-			qb_util_log(LOG_ERR,
-				    "Error adding socket to mainloop.");
-		}
 	}
 
 send_response:
@@ -662,19 +763,24 @@ send_response:
 	}
 
 	if (res == 0) {
-		qb_ipcs_connection_ref(c);
-		if (s->serv_fns.connection_created) {
-			s->serv_fns.connection_created(c);
+		if (s->type != QB_IPC_SOCKET) {
+			qb_ipcs_connection_ref(c);
+			if (s->serv_fns.connection_created) {
+				s->serv_fns.connection_created(c);
+			}
+			if (c->state == QB_IPCS_CONNECTION_ACTIVE) {
+				c->state = QB_IPCS_CONNECTION_ESTABLISHED;
+			}
+			qb_ipcs_connection_unref(c);
 		}
-		if (c->state == QB_IPCS_CONNECTION_ACTIVE) {
-			c->state = QB_IPCS_CONNECTION_ESTABLISHED;
-		}
-		qb_ipcs_connection_unref(c);
 	} else {
 		if (res == -EACCES) {
-			qb_util_log(LOG_ERR, "Invalid IPC credentials.");
+			qb_util_log(LOG_ERR, "Invalid IPC credentials (%s).",
+				    c->description);
 		} else {
-			qb_util_perror(LOG_ERR, "Error in connection setup");
+			errno = -res;
+			qb_util_perror(LOG_ERR, "Error in connection setup (%s)",
+				       c->description);
 		}
 		qb_ipcs_disconnect(c);
 	}
@@ -688,7 +794,32 @@ handle_connection_new_sock(struct qb_ipcs_service *s, int32_t sock, void *msg)
 	struct qb_ipc_event_connection_request *req = msg;
 
 	c = (struct qb_ipcs_connection *)req->connection;
+	qb_ipcs_connection_ref(c);
 	c->event.u.us.sock = sock;
+	if (c->state == QB_IPCS_CONNECTION_ACTIVE) {
+		c->state = QB_IPCS_CONNECTION_ESTABLISHED;
+	}
+	if (s->serv_fns.connection_created) {
+		s->serv_fns.connection_created(c);
+	}
+
+	if (c->state == QB_IPCS_CONNECTION_ESTABLISHED &&
+	    s->type == QB_IPC_SOCKET) {
+		int32_t res;
+		qb_ipcs_connection_ref(c);
+		res = s->poll_fns.dispatch_add(s->poll_priority,
+					       c->request.u.us.sock,
+					       POLLIN | POLLPRI | POLLNVAL,
+					       c,
+					       qb_ipcs_dispatch_connection_request);
+		if (res < 0) {
+			qb_util_log(LOG_ERR,
+				    "Error adding socket to mainloop (%s).",
+				    c->description);
+		}
+	}
+
+	qb_ipcs_connection_unref(c);
 }
 
 static int32_t
@@ -699,28 +830,29 @@ qb_ipcs_uc_recv_and_auth(int32_t sock, void *msg, size_t len,
 	struct msghdr msg_recv;
 	struct iovec iov_recv;
 
-#ifdef QB_LINUX
+#ifdef SO_PASSCRED
 	char cmsg_cred[CMSG_SPACE(sizeof(struct ucred))];
 	int off = 0;
 	int on = 1;
 #endif
-	msg_recv.msg_flags = 0;
 	msg_recv.msg_iov = &iov_recv;
 	msg_recv.msg_iovlen = 1;
 	msg_recv.msg_name = 0;
 	msg_recv.msg_namelen = 0;
-#ifdef QB_LINUX
+#ifdef SO_PASSCRED
 	msg_recv.msg_control = (void *)cmsg_cred;
 	msg_recv.msg_controllen = sizeof(cmsg_cred);
 #endif
 #ifdef QB_SOLARIS
 	msg_recv.msg_accrights = 0;
 	msg_recv.msg_accrightslen = 0;
+#else
+	msg_recv.msg_flags = 0;
 #endif /* QB_SOLARIS */
 
 	iov_recv.iov_base = msg;
 	iov_recv.iov_len = len;
-#ifdef QB_LINUX
+#ifdef SO_PASSCRED
 	setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
 #endif
 
@@ -776,16 +908,19 @@ qb_ipcs_uc_recv_and_auth(int32_t sock, void *msg, size_t len,
 	 */
 	{
 		struct ucred cred;
-		struct cmsghdr *cmsg = CMSG_FIRSTHDR(&msg_recv);
-		assert(cmsg != NULL);
-		if (CMSG_DATA(cmsg)) {
+		struct cmsghdr *cmsg;
+
+		res = -EINVAL;
+		for (cmsg = CMSG_FIRSTHDR(&msg_recv); cmsg != NULL; cmsg = CMSG_NXTHDR(&msg_recv, cmsg)) {
+			if (cmsg->cmsg_type != SCM_CREDENTIALS)
+				continue;
+
 			memcpy(&cred, CMSG_DATA(cmsg), sizeof(struct ucred));
 			res = 0;
 			ugp->pid = cred.pid;
 			ugp->uid = cred.uid;
 			ugp->gid = cred.gid;
-		} else {
-			res = -EBADMSG;
+			break;
 		}
 	}
 #else /* no credentials */
@@ -797,7 +932,7 @@ qb_ipcs_uc_recv_and_auth(int32_t sock, void *msg, size_t len,
 
 cleanup_and_return:
 
-#ifdef QB_LINUX
+#ifdef SO_PASSCRED
 	setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &off, sizeof(off));
 #endif
 
@@ -815,6 +950,13 @@ qb_ipcs_us_connection_acceptor(int fd, int revent, void *data)
 	struct ipc_auth_ugp ugp;
 	socklen_t addrlen = sizeof(struct sockaddr_un);
 
+	if (revent & (POLLNVAL|POLLHUP|POLLERR)) {
+		/*
+		 * handle shutdown more cleanly.
+		 */
+		return -1;
+	}
+
 retry_accept:
 	errno = 0;
 	new_fd = accept(fd, (struct sockaddr *)&un_addr, &addrlen);
@@ -847,6 +989,13 @@ retry_accept:
 
 	res = qb_ipcs_uc_recv_and_auth(new_fd, &setup_msg, sizeof(setup_msg),
 				       &ugp);
+	if (res < 0) {
+		close(new_fd);
+		/* This is an error, but -1 would indicate disconnect
+		 * from the poll loop
+		 */
+		return 0;
+	}
 
 	if (setup_msg.hdr.id == QB_IPC_MSG_AUTHENTICATE) {
 		(void)handle_new_connection(s, res, new_fd, &setup_msg,
@@ -873,37 +1022,60 @@ qb_ipcs_us_connect(struct qb_ipcs_service *s,
 	int32_t fd_hdr;
 	int32_t res = 0;
 	struct ipc_us_control *ctl;
+	char * shm_ptr;
 
-	qb_util_log(LOG_DEBUG, "connecting to client [%d]", c->pid);
+	qb_util_log(LOG_DEBUG, "connecting to client (%s)",
+		    c->description);
 
-	snprintf(r->request, NAME_MAX, "qb-%s-control-%d-%d",
-		 s->name, c->pid, c->setup.u.us.sock);
+	snprintf(r->request, NAME_MAX, "qb-%s-control-%s",
+		 s->name, c->description);
 
 	fd_hdr = qb_sys_mmap_file_open(path, r->request,
-				       sizeof(struct ipc_us_control),
+				       SHM_CONTROL_SIZE,
 				       O_CREAT | O_TRUNC | O_RDWR);
 	if (fd_hdr < 0) {
 		res = -errno;
-		qb_util_perror(LOG_ERR, "couldn't create file for mmap");
+		qb_util_perror(LOG_ERR, "couldn't create file for mmap (%s)",
+			       c->description);
 		return res;
 	}
 	(void)strlcpy(r->request, path, PATH_MAX);
 	(void)strlcpy(c->request.u.us.shared_file_name, r->request, NAME_MAX);
-
-	c->request.u.us.shared_data = mmap(0,
-					   sizeof(struct ipc_us_control),
-					   PROT_READ | PROT_WRITE, MAP_SHARED,
-					   fd_hdr, 0);
-
-	if (c->request.u.us.shared_data == MAP_FAILED) {
-		res = -errno;
-		qb_util_perror(LOG_ERR, "couldn't create mmap for header");
-		goto cleanup_hdr;
+	res = chown(r->request, c->auth.uid, c->auth.gid);
+	if (res != 0) {
+		/* ignore res, this is just for the compiler warnings.
+		 */
+		res = 0;
+	}
+	res = chmod(r->request, c->auth.mode);
+	if (res != 0) {
+		/* ignore res, this is just for the compiler warnings.
+		*/
+		res = 0;
 	}
 
+	shm_ptr = mmap(0, SHM_CONTROL_SIZE,
+		       PROT_READ | PROT_WRITE, MAP_SHARED, fd_hdr, 0);
+
+	if (shm_ptr == MAP_FAILED) {
+		res = -errno;
+		qb_util_perror(LOG_ERR, "couldn't create mmap for header (%s)",
+			       c->description);
+		goto cleanup_hdr;
+	}
+	c->request.u.us.shared_data = shm_ptr;
+	c->response.u.us.shared_data = shm_ptr + sizeof(struct ipc_us_control);
+	c->event.u.us.shared_data =  shm_ptr + (2 * sizeof(struct ipc_us_control));
+
 	ctl = (struct ipc_us_control *)c->request.u.us.shared_data;
 	ctl->sent = 0;
 	ctl->flow_control = 0;
+	ctl = (struct ipc_us_control *)c->response.u.us.shared_data;
+	ctl->sent = 0;
+	ctl->flow_control = 0;
+	ctl = (struct ipc_us_control *)c->event.u.us.shared_data;
+	ctl->sent = 0;
+	ctl->flow_control = 0;
 
 	close(fd_hdr);
 	return res;
@@ -911,7 +1083,7 @@ qb_ipcs_us_connect(struct qb_ipcs_service *s,
 cleanup_hdr:
 	close(fd_hdr);
 	unlink(r->request);
-	munmap(c->request.u.us.shared_data, sizeof(struct ipc_us_control));
+	munmap(c->request.u.us.shared_data, SHM_CONTROL_SIZE);
 	return res;
 }
 
@@ -921,7 +1093,8 @@ qb_ipc_us_fc_set(struct qb_ipc_one_way *one_way, int32_t fc_enable)
 	struct ipc_us_control *ctl =
 	    (struct ipc_us_control *)one_way->u.us.shared_data;
 
-	qb_util_log(LOG_TRACE, "setting fc to %d", fc_enable);
+	qb_util_log(LOG_TRACE, "setting fc to %d",
+		    fc_enable);
 	qb_atomic_int_set(&ctl->flow_control, fc_enable);
 }
 
@@ -942,14 +1115,35 @@ qb_ipc_us_q_len_get(struct qb_ipc_one_way *one_way)
 	return qb_atomic_int_get(&ctl->sent);
 }
 
+void
+qb_ipcs_sockets_disconnect(struct qb_ipcs_connection *c)
+{
+	int sock = -1;
+
+	qb_enter();
+	if (c->service->needs_sock_for_poll && c->setup.u.us.sock > 0) {
+		sock = c->setup.u.us.sock;
+		qb_ipcc_us_sock_close(sock);
+		c->setup.u.us.sock = -1;
+	}
+	if (c->request.type == QB_IPC_SOCKET) {
+		sock = c->request.u.us.sock;
+	}
+	if (sock > 0) {
+		(void)c->service->poll_fns.dispatch_del(sock);
+		qb_ipcs_connection_unref(c);
+	}
+}
+
 static void
 qb_ipcs_us_disconnect(struct qb_ipcs_connection *c)
 {
-	munmap(c->request.u.us.shared_data, sizeof(struct ipc_us_control));
+	qb_enter();
+	munmap(c->request.u.us.shared_data, SHM_CONTROL_SIZE);
 	unlink(c->request.u.us.shared_file_name);
 
-	close(c->request.u.us.sock);
-	close(c->event.u.us.sock);
+	qb_ipcc_us_sock_close(c->request.u.us.sock);
+	qb_ipcc_us_sock_close(c->event.u.us.sock);
 }
 
 void
@@ -968,4 +1162,6 @@ qb_ipcs_us_init(struct qb_ipcs_service *s)
 	s->funcs.q_len_get = qb_ipc_us_q_len_get;
 
 	s->needs_sock_for_poll = QB_FALSE;
+
+	qb_atomic_init();
 }
