@@ -225,7 +225,7 @@ qb_log_thread_log_write(struct qb_log_callsite *cs,
 		if (t->state != QB_LOG_STATE_ENABLED) {
 			continue;
 		}
-		if (t->threaded) {
+		if (!t->threaded) {
 			continue;
 		}
 		if (qb_bit_is_set(cs->targets, t->pos)) {
@@ -287,6 +287,11 @@ qb_log_callsite_get(const char *function,
 			_custom_filter_fn(cs);
 		}
 		pthread_rwlock_unlock(&_listlock);
+	} else if (cs->tags != tags) {
+		cs->tags = tags;
+		if (_custom_filter_fn) {
+			_custom_filter_fn(cs);
+		}
 	}
 	return cs;
 }
@@ -386,21 +391,35 @@ qb_log_callsites_register(struct qb_log_callsite *_start,
 	pthread_rwlock_unlock(&_listlock);
 	if (_custom_filter_fn) {
 		for (cs = sect->start; cs < sect->stop; cs++) {
-			if (cs->lineno == 0) {
-				break;
+			if (cs->lineno > 0) {
+				_custom_filter_fn(cs);
 			}
-			_custom_filter_fn(cs);
 		}
 	}
+	/* qb_log_callsites_dump_sect(sect); */
 
 	return 0;
 }
 
+static void
+qb_log_callsites_dump_sect(struct callsite_section *sect)
+{
+	struct qb_log_callsite *cs;
+
+	printf(" start %p - stop %p\n", sect->start, sect->stop);
+	printf("filename    lineno targets         tags\n");
+	for (cs = sect->start; cs < sect->stop; cs++) {
+		if (cs->lineno > 0) {
+			printf("%12s %6d %16d %16d\n", cs->filename, cs->lineno,
+			       cs->targets, cs->tags);
+		}
+	}
+}
+
 void
 qb_log_callsites_dump(void)
 {
 	struct callsite_section *sect;
-	struct qb_log_callsite *cs;
 	int32_t l;
 
 	pthread_rwlock_rdlock(&_listlock);
@@ -408,15 +427,7 @@ qb_log_callsites_dump(void)
 	printf("Callsite Database [%d]\n", l);
 	printf("---------------------\n");
 	qb_list_for_each_entry(sect, &callsite_sections, list) {
-		printf(" start %p - stop %p\n", sect->start, sect->stop);
-		printf("filename    lineno targets         tags\n");
-		for (cs = sect->start; cs < sect->stop; cs++) {
-			if (cs->lineno == 0) {
-				break;
-			}
-			printf("%12s %6d %16d %16d\n", cs->filename, cs->lineno,
-			       cs->targets, cs->tags);
-		}
+		qb_log_callsites_dump_sect(sect);
 	}
 	pthread_rwlock_unlock(&_listlock);
 }
@@ -531,11 +542,10 @@ _log_filter_apply(struct callsite_section *sect,
 	struct qb_log_callsite *cs;
 
 	for (cs = sect->start; cs < sect->stop; cs++) {
-		if (cs->lineno == 0) {
-			break;
+		if (cs->lineno > 0) {
+			_log_filter_apply_to_cs(cs, t, c, type, text,
+					    high_priority, low_priority);
 		}
-		_log_filter_apply_to_cs(cs, t, c, type, text,
-					high_priority, low_priority);
 	}
 }
 
@@ -641,10 +651,9 @@ qb_log_filter_fn_set(qb_log_filter_fn fn)
 
 	qb_list_for_each_entry(sect, &callsite_sections, list) {
 		for (cs = sect->start; cs < sect->stop; cs++) {
-			if (cs->lineno == 0) {
-				break;
+			if (cs->lineno > 0) {
+				_custom_filter_fn(cs);
 			}
-			_custom_filter_fn(cs);
 		}
 	}
 	return 0;
@@ -730,6 +739,7 @@ qb_log_init(const char *name, int32_t facility, uint8_t priority)
 	for (i = 0; i < QB_LOG_TARGET_MAX; i++) {
 		conf[i].pos = i;
 		conf[i].debug = QB_FALSE;
+		conf[i].file_sync = QB_FALSE;
 		conf[i].state = QB_LOG_STATE_UNUSED;
 		(void)strlcpy(conf[i].name, name, PATH_MAX);
 		conf[i].facility = facility;
@@ -894,7 +904,9 @@ qb_log_custom_close(int32_t t)
 	target = qb_log_target_get(t);
 
 	if (target->close) {
+		in_logger = QB_TRUE;
 		target->close(t);
+		in_logger = QB_FALSE;
 	}
 	qb_log_target_free(target);
 }
@@ -965,6 +977,9 @@ qb_log_ctl(int32_t t, enum qb_log_conf c, int32_t arg)
 			need_reload = QB_TRUE;
 		}
 		break;
+	case QB_LOG_CONF_FILE_SYNC:
+		conf[t].file_sync = arg;
+		break;
 	case QB_LOG_CONF_PRIORITY_BUMP:
 		conf[t].priority_bump = arg;
 		break;
@@ -987,7 +1002,9 @@ qb_log_ctl(int32_t t, enum qb_log_conf c, int32_t arg)
 		rc = -EINVAL;
 	}
 	if (rc == 0 && need_reload && conf[t].reload) {
+		in_logger = QB_TRUE;
 		conf[t].reload(t);
+		in_logger = QB_FALSE;
 	}
 	return rc;
 }
