@@ -21,12 +21,8 @@
  * along with libqb.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <sys/types.h>
+#include "os_base.h"
 #include <sys/wait.h>
-#include <errno.h>
 #include <signal.h>
 #include <check.h>
 
@@ -46,8 +42,12 @@ enum my_msg_ids {
 	IPC_MSG_RES_TX_RX,
 	IPC_MSG_REQ_DISPATCH,
 	IPC_MSG_RES_DISPATCH,
+	IPC_MSG_REQ_BULK_EVENTS,
+	IPC_MSG_RES_BULK_EVENTS,
 	IPC_MSG_REQ_SERVER_FAIL,
 	IPC_MSG_RES_SERVER_FAIL,
+	IPC_MSG_REQ_SERVER_DISCONNECT,
+	IPC_MSG_RES_SERVER_DISCONNECT,
 };
 
 /* Test Cases
@@ -72,6 +72,9 @@ static qb_loop_t *my_loop;
 static qb_ipcs_service_t* s1;
 static int32_t turn_on_fc = QB_FALSE;
 static int32_t fc_enabled = 89;
+static int32_t send_event_on_created = QB_FALSE;
+static int32_t disconnect_after_created = QB_FALSE;
+static int32_t num_bulk_events = 10;
 
 static int32_t
 exit_handler(int32_t rsignal, void *data)
@@ -108,12 +111,40 @@ s1_msg_process_fn(qb_ipcs_connection_t *c,
 		response.id = IPC_MSG_RES_DISPATCH;
 		response.error = 0;
 		res = qb_ipcs_event_send(c, &response,
-				sizeof(response));
+					 sizeof(response));
 		if (res < 0) {
 			qb_perror(LOG_INFO, "qb_ipcs_event_send");
 		}
+	} else if (req_pt->id == IPC_MSG_REQ_BULK_EVENTS) {
+		int32_t m;
+		int32_t num;
+		struct qb_ipcs_connection_stats_2 *stats;
+
+		response.size = sizeof(struct qb_ipc_response_header);
+		response.error = 0;
+
+		stats = qb_ipcs_connection_stats_get_2(c, QB_FALSE);
+		num = stats->event_q_length;
+		free(stats);
+
+		for (m = 0; m < num_bulk_events; m++) {
+			res = qb_ipcs_event_send(c, &response,
+						 sizeof(response));
+			ck_assert_int_eq(res, sizeof(response));
+			response.id++;
+		}
+		stats = qb_ipcs_connection_stats_get_2(c, QB_FALSE);
+		ck_assert_int_eq(stats->event_q_length - num, num_bulk_events);
+		free(stats);
+
+		response.id = IPC_MSG_RES_BULK_EVENTS;
+		res = qb_ipcs_response_send(c, &response, response.size);
+		ck_assert_int_eq(res, sizeof(response));
+
 	} else if (req_pt->id == IPC_MSG_REQ_SERVER_FAIL) {
 		exit(0);
+	} else if (req_pt->id == IPC_MSG_REQ_SERVER_DISCONNECT) {
+		qb_ipcs_disconnect(c);
 	}
 	return 0;
 }
@@ -146,11 +177,36 @@ my_dispatch_del(int32_t fd)
 	return qb_loop_poll_del(my_loop, fd);
 }
 
+static int32_t
+s1_connection_closed(qb_ipcs_connection_t *c)
+{
+	qb_enter();
+	qb_leave();
+	return 0;
+}
+
 static void
 s1_connection_destroyed(qb_ipcs_connection_t *c)
 {
 	qb_enter();
 	qb_loop_stop(my_loop);
+	qb_leave();
+}
+
+static void
+s1_connection_created(qb_ipcs_connection_t *c)
+{
+	if (send_event_on_created) {
+		struct qb_ipc_response_header response;
+		int32_t res;
+
+		response.size = sizeof(struct qb_ipc_response_header);
+		response.id = IPC_MSG_RES_DISPATCH;
+		response.error = 0;
+		res = qb_ipcs_event_send(c, &response,
+					 sizeof(response));
+		ck_assert_int_eq(res, response.size);
+	}
 }
 
 static void
@@ -161,10 +217,10 @@ run_ipc_server(void)
 
 	struct qb_ipcs_service_handlers sh = {
 		.connection_accept = NULL,
-		.connection_created = NULL,
+		.connection_created = s1_connection_created,
 		.msg_process = s1_msg_process_fn,
 		.connection_destroyed = s1_connection_destroyed,
-		.connection_closed = NULL,
+		.connection_closed = s1_connection_closed,
 	};
 
 	struct qb_ipcs_poll_handlers ph = {
@@ -226,13 +282,14 @@ struct my_req {
 
 static struct my_req request;
 static int32_t
-send_and_check(uint32_t size, int32_t ms_timeout, int32_t expect_perfection)
+send_and_check(int32_t req_id, uint32_t size,
+	       int32_t ms_timeout, int32_t expect_perfection)
 {
 	struct qb_ipc_response_header res_header;
 	int32_t res;
 	int32_t try_times = 0;
 
-	request.hdr.id = IPC_MSG_REQ_TX_RX;
+	request.hdr.id = req_id;
 	request.hdr.size = sizeof(struct qb_ipc_request_header) + size;
 
 repeat_send:
@@ -251,8 +308,16 @@ repeat_send:
 			return res;
 		}
 	}
-	res = qb_ipcc_recv(conn, &res_header,
-			sizeof(struct qb_ipc_response_header), ms_timeout);
+
+	if (req_id == IPC_MSG_REQ_DISPATCH) {
+		res = qb_ipcc_event_recv(conn, &res_header,
+					 sizeof(struct qb_ipc_response_header),
+					 ms_timeout);
+	} else {
+		res = qb_ipcc_recv(conn, &res_header,
+				   sizeof(struct qb_ipc_response_header),
+				   ms_timeout);
+	}
 	if (res == -EINTR) {
 		return -1;
 	}
@@ -263,7 +328,7 @@ repeat_send:
 	}
 	if (expect_perfection) {
 		ck_assert_int_eq(res, sizeof(struct qb_ipc_response_header));
-		ck_assert_int_eq(res_header.id, IPC_MSG_RES_TX_RX);
+		ck_assert_int_eq(res_header.id, req_id + 1);
 		ck_assert_int_eq(res_header.size, sizeof(struct qb_ipc_response_header));
 	}
 	return res;
@@ -298,7 +363,8 @@ test_ipc_txrx(void)
 		size *= 2;
 		if (size >= MAX_MSG_SIZE)
 			break;
-		if (send_and_check(size, recv_timeout, QB_TRUE) < 0) {
+		if (send_and_check(IPC_MSG_REQ_TX_RX, size,
+				   recv_timeout, QB_TRUE) < 0) {
 			break;
 		}
 	}
@@ -312,9 +378,12 @@ test_ipc_txrx(void)
 static void
 test_ipc_exit(void)
 {
+	struct qb_ipc_request_header req_header;
+	struct qb_ipc_response_header res_header;
+	struct iovec iov[1];
+	int32_t res;
 	int32_t c = 0;
 	int32_t j = 0;
-	int32_t rc = 0;
 	pid_t pid;
 
 	pid = run_function_in_new_process(run_ipc_server);
@@ -332,9 +401,16 @@ test_ipc_exit(void)
 	} while (conn == NULL && c < 5);
 	fail_if(conn == NULL);
 
-	/* confirm we can send one message */
-	rc = send_and_check(100, recv_timeout, QB_TRUE);
-	_ck_assert_int(rc, >=, 0);
+	req_header.id = IPC_MSG_REQ_TX_RX;
+	req_header.size = sizeof(struct qb_ipc_request_header);
+
+	iov[0].iov_len = req_header.size;
+	iov[0].iov_base = &req_header;
+
+	res = qb_ipcc_sendv_recv(conn, iov, 1,
+				 &res_header,
+				 sizeof(struct qb_ipc_response_header), -1);
+	ck_assert_int_eq(res, sizeof(struct qb_ipc_response_header));
 
 	/* kill the server */
 	stop_process(pid);
@@ -350,91 +426,97 @@ test_ipc_exit(void)
 	qb_ipcc_disconnect(conn);
 }
 
+START_TEST(test_ipc_exit_us)
+{
+	qb_enter();
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	recv_timeout = 5000;
+	test_ipc_exit();
+	qb_leave();
+}
+END_TEST
 
 START_TEST(test_ipc_exit_shm)
 {
+	qb_enter();
 	ipc_type = QB_IPC_SHM;
 	ipc_name = __func__;
 	recv_timeout = 1000;
 	test_ipc_exit();
+	qb_leave();
 }
 END_TEST
 
 START_TEST(test_ipc_txrx_shm_tmo)
 {
+	qb_enter();
 	turn_on_fc = QB_FALSE;
 	ipc_type = QB_IPC_SHM;
 	ipc_name = __func__;
 	recv_timeout = 1000;
 	test_ipc_txrx();
+	qb_leave();
 }
 END_TEST
 
 START_TEST(test_ipc_txrx_shm_block)
 {
+	qb_enter();
 	turn_on_fc = QB_FALSE;
 	ipc_type = QB_IPC_SHM;
 	ipc_name = __func__;
 	recv_timeout = -1;
 	test_ipc_txrx();
-}
-END_TEST
-
-START_TEST(test_ipc_txrx_us_block)
-{
-	turn_on_fc = QB_FALSE;
-	ipc_type = QB_IPC_SOCKET;
-	ipc_name = __func__;
-	recv_timeout = -1;
-	test_ipc_txrx();
-}
-END_TEST
-
-START_TEST(test_ipc_txrx_us_tmo)
-{
-	turn_on_fc = QB_FALSE;
-	ipc_type = QB_IPC_SOCKET;
-	ipc_name = __func__;
-	recv_timeout = 1000;
-	test_ipc_txrx();
+	qb_leave();
 }
 END_TEST
 
 START_TEST(test_ipc_fc_shm)
 {
+	qb_enter();
 	turn_on_fc = QB_TRUE;
 	ipc_type = QB_IPC_SHM;
 	recv_timeout = 500;
 	ipc_name = __func__;
 	test_ipc_txrx();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_txrx_us_block)
+{
+	qb_enter();
+	turn_on_fc = QB_FALSE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	recv_timeout = -1;
+	test_ipc_txrx();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_txrx_us_tmo)
+{
+	qb_enter();
+	turn_on_fc = QB_FALSE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	recv_timeout = 1000;
+	test_ipc_txrx();
+	qb_leave();
 }
 END_TEST
 
 START_TEST(test_ipc_fc_us)
 {
+	qb_enter();
 	turn_on_fc = QB_TRUE;
 	ipc_type = QB_IPC_SOCKET;
 	recv_timeout = 500;
 	ipc_name = __func__;
 	test_ipc_txrx();
-}
-END_TEST
-
-START_TEST(test_ipc_txrx_pmq)
-{
-	turn_on_fc = QB_FALSE;
-	ipc_type = QB_IPC_POSIX_MQ;
-	ipc_name = __func__;
-	test_ipc_txrx();
-}
-END_TEST
-
-START_TEST(test_ipc_txrx_smq)
-{
-	turn_on_fc = QB_FALSE;
-	ipc_type = QB_IPC_SYSV_MQ;
-	ipc_name = __func__;
-	test_ipc_txrx();
+	qb_leave();
 }
 END_TEST
 
@@ -443,15 +525,13 @@ struct my_res {
 	char message[1024 * 1024];
 };
 
-static struct my_res response;
 static void
 test_ipc_dispatch(void)
 {
-	int32_t res;
 	int32_t j;
 	int32_t c = 0;
 	pid_t pid;
-	struct qb_ipc_request_header req_header;
+	int32_t size;
 
 	pid = run_function_in_new_process(run_ipc_server);
 	fail_if(pid == -1);
@@ -468,54 +548,230 @@ test_ipc_dispatch(void)
 	} while (conn == NULL && c < 5);
 	fail_if(conn == NULL);
 
-	req_header.id = IPC_MSG_REQ_DISPATCH;
-	req_header.size = sizeof(struct qb_ipc_request_header);
-
- repeat_send:
-	res = qb_ipcc_send(conn, &req_header, req_header.size);
-	if (res < 0) {
-		if (res == -EAGAIN) {
-			goto repeat_send;
-		} else if (res == -EINVAL || res == -EINTR) {
-			qb_perror(LOG_INFO, "qb_ipcc_send");
-			return;
-		} else {
-			errno = -res;
-			qb_perror(LOG_INFO, "qb_ipcc_send");
-			goto repeat_send;
+	size = QB_MIN(sizeof(struct qb_ipc_request_header), 64);
+	for (j = 1; j < 19; j++) {
+		size *= 2;
+		if (size >= MAX_MSG_SIZE)
+			break;
+		if (send_and_check(IPC_MSG_REQ_DISPATCH, size,
+				   recv_timeout, QB_TRUE) < 0) {
+			break;
 		}
 	}
- repeat_event_recv:
-	res = qb_ipcc_event_recv(conn, &response, sizeof(struct my_res), 0);
-	if (res < 0) {
-		if (res == -EAGAIN || res == -ETIMEDOUT) {
-			goto repeat_event_recv;
-		} else {
-			errno = -res;
-			qb_perror(LOG_INFO, "qb_ipcc_event_recv");
-			goto repeat_send;
-		}
-	}
-	ck_assert_int_eq(res, sizeof(struct qb_ipc_response_header));
-	ck_assert_int_eq(response.hdr.id, IPC_MSG_RES_DISPATCH);
 
 	qb_ipcc_disconnect(conn);
 	stop_process(pid);
 }
 
-START_TEST(test_ipc_disp_shm)
-{
-	ipc_type = QB_IPC_SHM;
-	ipc_name = __func__;
-	test_ipc_dispatch();
-}
-END_TEST
-
 START_TEST(test_ipc_disp_us)
 {
+	qb_enter();
 	ipc_type = QB_IPC_SOCKET;
 	ipc_name = __func__;
 	test_ipc_dispatch();
+	qb_leave();
+}
+END_TEST
+
+static int32_t events_received;
+
+
+static int32_t
+count_bulk_events(int32_t fd, int32_t revents, void *data)
+{
+	qb_loop_t *cl = (qb_loop_t*)data;
+
+	events_received++;
+
+	if (events_received >= num_bulk_events) {
+		qb_loop_stop(cl);
+		return -1;
+	}
+	return 0;
+}
+
+static void
+test_ipc_bulk_events(void)
+{
+	struct qb_ipc_request_header req_header;
+	struct qb_ipc_response_header res_header;
+	struct iovec iov[1];
+	int32_t c = 0;
+	int32_t j = 0;
+	pid_t pid;
+	int32_t res;
+	qb_loop_t *cl;
+	int32_t fd;
+
+	pid = run_function_in_new_process(run_ipc_server);
+	fail_if(pid == -1);
+	sleep(1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, MAX_MSG_SIZE);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			sleep(1);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	events_received = 0;
+	cl = qb_loop_create();
+	res = qb_ipcc_fd_get(conn, &fd),
+	ck_assert_int_eq(res, 0);
+	res = qb_loop_poll_add(cl, QB_LOOP_MED,
+			 fd, POLLIN,
+			 cl, count_bulk_events);
+	ck_assert_int_eq(res, 0);
+
+	res = send_and_check(IPC_MSG_REQ_BULK_EVENTS,
+			     0,
+			     recv_timeout, QB_TRUE);
+	ck_assert_int_eq(res, sizeof(struct qb_ipc_response_header));
+
+	qb_loop_run(cl);
+	ck_assert_int_eq(events_received, num_bulk_events);
+
+	req_header.id = IPC_MSG_REQ_SERVER_FAIL;
+	req_header.size = sizeof(struct qb_ipc_request_header);
+
+	iov[0].iov_len = req_header.size;
+	iov[0].iov_base = &req_header;
+	res = qb_ipcc_sendv_recv(conn, iov, 1,
+				 &res_header,
+				 sizeof(struct qb_ipc_response_header), -1);
+	if (res != -ECONNRESET && res != -ENOTCONN) {
+		qb_log(LOG_ERR, "id:%d size:%d", res_header.id, res_header.size);
+		ck_assert_int_eq(res, -ENOTCONN);
+	}
+
+	qb_ipcc_disconnect(conn);
+	stop_process(pid);
+}
+
+START_TEST(test_ipc_bulk_events_us)
+{
+	qb_enter();
+	send_event_on_created = QB_FALSE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	test_ipc_bulk_events();
+	qb_leave();
+}
+END_TEST
+
+static void
+test_ipc_event_on_created(void)
+{
+	int32_t c = 0;
+	int32_t j = 0;
+	pid_t pid;
+	int32_t res;
+	qb_loop_t *cl;
+	int32_t fd;
+
+	num_bulk_events = 1;
+
+	pid = run_function_in_new_process(run_ipc_server);
+	fail_if(pid == -1);
+	sleep(1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, MAX_MSG_SIZE);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			sleep(1);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	events_received = 0;
+	cl = qb_loop_create();
+	res = qb_ipcc_fd_get(conn, &fd),
+	ck_assert_int_eq(res, 0);
+	res = qb_loop_poll_add(cl, QB_LOOP_MED,
+			 fd, POLLIN,
+			 cl, count_bulk_events);
+	ck_assert_int_eq(res, 0);
+
+	qb_loop_run(cl);
+	ck_assert_int_eq(events_received, num_bulk_events);
+
+	qb_ipcc_disconnect(conn);
+	stop_process(pid);
+}
+
+START_TEST(test_ipc_event_on_created_us)
+{
+	qb_enter();
+	send_event_on_created = QB_TRUE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	test_ipc_event_on_created();
+	qb_leave();
+}
+END_TEST
+
+static void
+test_ipc_disconnect_after_created(void)
+{
+	struct qb_ipc_request_header req_header;
+	struct qb_ipc_response_header res_header;
+	struct iovec iov[1];
+	int32_t c = 0;
+	int32_t j = 0;
+	pid_t pid;
+	int32_t res;
+
+	pid = run_function_in_new_process(run_ipc_server);
+	fail_if(pid == -1);
+	sleep(1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, MAX_MSG_SIZE);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			sleep(1);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	ck_assert_int_eq(QB_TRUE, qb_ipcc_is_connected(conn));
+
+	req_header.id = IPC_MSG_REQ_SERVER_DISCONNECT;
+	req_header.size = sizeof(struct qb_ipc_request_header);
+
+	iov[0].iov_len = req_header.size;
+	iov[0].iov_base = &req_header;
+
+	res = qb_ipcc_sendv_recv(conn, iov, 1,
+				 &res_header,
+				 sizeof(struct qb_ipc_response_header), -1);
+	/*
+	 * confirm we get -ENOTCONN
+	 */
+	ck_assert_int_eq(res, -ENOTCONN);
+	ck_assert_int_eq(QB_FALSE, qb_ipcc_is_connected(conn));
+
+	qb_ipcc_disconnect(conn);
+	stop_process(pid);
+}
+
+START_TEST(test_ipc_disconnect_after_created_us)
+{
+	qb_enter();
+	disconnect_after_created = QB_TRUE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	test_ipc_disconnect_after_created();
+	qb_leave();
 }
 END_TEST
 
@@ -554,6 +810,8 @@ test_ipc_server_fail(void)
 	iov[0].iov_len = req_header.size;
 	iov[0].iov_base = &req_header;
 
+	ck_assert_int_eq(QB_TRUE, qb_ipcc_is_connected(conn));
+
 	res = qb_ipcc_sendv_recv(conn, iov, 1,
 				 &res_header,
 				 sizeof(struct qb_ipc_response_header), -1);
@@ -561,6 +819,7 @@ test_ipc_server_fail(void)
 	 * confirm we get -ENOTCONN
 	 */
 	ck_assert_int_eq(res, -ENOTCONN);
+	ck_assert_int_eq(QB_FALSE, qb_ipcc_is_connected(conn));
 
 	qb_ipcc_disconnect(conn);
 	stop_process(pid);
@@ -568,37 +827,67 @@ test_ipc_server_fail(void)
 
 START_TEST(test_ipc_server_fail_soc)
 {
+	qb_enter();
 	ipc_type = QB_IPC_SOCKET;
 	ipc_name = __func__;
 	test_ipc_server_fail();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_disp_shm)
+{
+	qb_enter();
+	ipc_type = QB_IPC_SHM;
+	ipc_name = __func__;
+	test_ipc_dispatch();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_bulk_events_shm)
+{
+	qb_enter();
+	ipc_type = QB_IPC_SHM;
+	ipc_name = __func__;
+	test_ipc_bulk_events();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_event_on_created_shm)
+{
+	qb_enter();
+	send_event_on_created = QB_TRUE;
+	ipc_type = QB_IPC_SHM;
+	ipc_name = __func__;
+	test_ipc_event_on_created();
+	qb_leave();
 }
 END_TEST
 
 START_TEST(test_ipc_server_fail_shm)
 {
+	qb_enter();
 	ipc_type = QB_IPC_SHM;
 	ipc_name = __func__;
 	test_ipc_server_fail();
+	qb_leave();
 }
 END_TEST
 
+
 static Suite *
-ipc_suite(void)
+make_shm_suite(void)
 {
 	TCase *tc;
-	uid_t uid;
-	Suite *s = suite_create("ipc");
+	Suite *s = suite_create("shm");
 
 	tc = tcase_create("ipc_server_fail_shm");
 	tcase_add_test(tc, test_ipc_server_fail_shm);
 	tcase_set_timeout(tc, 6);
 	suite_add_tcase(s, tc);
 
-	tc = tcase_create("ipc_server_fail_soc");
-	tcase_add_test(tc, test_ipc_server_fail_soc);
-	tcase_set_timeout(tc, 6);
-	suite_add_tcase(s, tc);
-
 	tc = tcase_create("ipc_txrx_shm_block");
 	tcase_add_test(tc, test_ipc_txrx_shm_block);
 	tcase_set_timeout(tc, 6);
@@ -609,6 +898,44 @@ ipc_suite(void)
 	tcase_set_timeout(tc, 6);
 	suite_add_tcase(s, tc);
 
+	tc = tcase_create("ipc_fc_shm");
+	tcase_add_test(tc, test_ipc_fc_shm);
+	tcase_set_timeout(tc, 6);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_dispatch_shm");
+	tcase_add_test(tc, test_ipc_disp_shm);
+	tcase_set_timeout(tc, 16);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_bulk_events_shm");
+	tcase_add_test(tc, test_ipc_bulk_events_shm);
+	tcase_set_timeout(tc, 16);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_exit_shm");
+	tcase_add_test(tc, test_ipc_exit_shm);
+	tcase_set_timeout(tc, 3);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_event_on_created_shm");
+	tcase_add_test(tc, test_ipc_event_on_created_shm);
+	suite_add_tcase(s, tc);
+
+	return s;
+}
+
+static Suite *
+make_soc_suite(void)
+{
+	Suite *s = suite_create("socket");
+	TCase *tc;
+
+	tc = tcase_create("ipc_server_fail_soc");
+	tcase_add_test(tc, test_ipc_server_fail_soc);
+	tcase_set_timeout(tc, 6);
+	suite_add_tcase(s, tc);
+
 	tc = tcase_create("ipc_txrx_us_block");
 	tcase_add_test(tc, test_ipc_txrx_us_block);
 	tcase_set_timeout(tc, 6);
@@ -619,31 +946,14 @@ ipc_suite(void)
 	tcase_set_timeout(tc, 6);
 	suite_add_tcase(s, tc);
 
-	tc = tcase_create("ipc_fc_shm");
-	tcase_add_test(tc, test_ipc_fc_shm);
-	tcase_set_timeout(tc, 6);
-	suite_add_tcase(s, tc);
-
 	tc = tcase_create("ipc_fc_us");
 	tcase_add_test(tc, test_ipc_fc_us);
 	tcase_set_timeout(tc, 6);
 	suite_add_tcase(s, tc);
 
-	uid = geteuid();
-	if (uid == 0) {
-		tc = tcase_create("ipc_txrx_posix_mq");
-		tcase_add_test(tc, test_ipc_txrx_pmq);
-		tcase_set_timeout(tc, 10);
-		suite_add_tcase(s, tc);
-
-		tc = tcase_create("ipc_txrx_sysv_mq");
-		tcase_add_test(tc, test_ipc_txrx_smq);
-		tcase_set_timeout(tc, 10);
-		suite_add_tcase(s, tc);
-	}
-	tc = tcase_create("ipc_dispatch_shm");
-	tcase_add_test(tc, test_ipc_disp_shm);
-	tcase_set_timeout(tc, 16);
+	tc = tcase_create("ipc_exit_us");
+	tcase_add_test(tc, test_ipc_exit_us);
+	tcase_set_timeout(tc, 6);
 	suite_add_tcase(s, tc);
 
 	tc = tcase_create("ipc_dispatch_us");
@@ -651,9 +961,17 @@ ipc_suite(void)
 	tcase_set_timeout(tc, 16);
 	suite_add_tcase(s, tc);
 
-	tc = tcase_create("ipc_exit_shm");
-	tcase_add_test(tc, test_ipc_exit_shm);
-	tcase_set_timeout(tc, 3);
+	tc = tcase_create("ipc_bulk_events_us");
+	tcase_add_test(tc, test_ipc_bulk_events_us);
+	tcase_set_timeout(tc, 16);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_event_on_created_us");
+	tcase_add_test(tc, test_ipc_event_on_created_us);
+	suite_add_tcase(s, tc);
+
+	tc = tcase_create("ipc_disconnect_after_created_us");
+	tcase_add_test(tc, test_ipc_disconnect_after_created_us);
 	suite_add_tcase(s, tc);
 
 	return s;
@@ -663,15 +981,27 @@ int32_t
 main(void)
 {
 	int32_t number_failed;
+	SRunner *sr;
+	Suite *s;
+	int32_t do_shm_tests = QB_TRUE;
 
-	Suite *s = ipc_suite();
-	SRunner *sr = srunner_create(s);
+#ifdef DISABLE_IPC_SHM
+	do_shm_tests = QB_FALSE;
+#endif /* DISABLE_IPC_SHM */
+
+	s = make_soc_suite();
+	sr = srunner_create(s);
+
+	if (do_shm_tests) {
+		srunner_add_suite(sr, make_shm_suite());
+	}
 
 	qb_log_init("check", LOG_USER, LOG_EMERG);
 	qb_log_ctl(QB_LOG_SYSLOG, QB_LOG_CONF_ENABLED, QB_FALSE);
 	qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD,
-			  QB_LOG_FILTER_FILE, "*", LOG_INFO);
+			  QB_LOG_FILTER_FILE, "*", LOG_TRACE);
 	qb_log_ctl(QB_LOG_STDERR, QB_LOG_CONF_ENABLED, QB_TRUE);
+	qb_log_format_set(QB_LOG_STDERR, "lib/%f|%l| %b");
 
 	srunner_run_all(sr, CK_VERBOSE);
 	number_failed = srunner_ntests_failed(sr);
